"""
ModelOutput class is used to build different output objects based on
user input. For now, available output types are pandas DataFrame or
netCDF4 Dataset.
The OutputHandlerInterface class is an interface for the creation of handlers
for other output object types.
"""
import abc
import time as t

from csv import QUOTE_NONE
from pathlib import Path
from collections import defaultdict

import regex as re

import numpy as np
import xarray as xr
import pandas as pd

from pysd._version import __version__
from pysd.tools.ncfiles import NCFile

from . utils import xrsplit


class OutputHandlerInterface(metaclass=abc.ABCMeta):
    """
    Interface for the creation of different output handlers.
    """

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'initialize') and
                callable(subclass.initialize) and
                hasattr(subclass, 'update') and
                callable(subclass.update) and
                hasattr(subclass, 'postprocess') and
                callable(subclass.postprocess) and
                hasattr(subclass, 'add_run_elements') and
                callable(subclass.add_run_elements) or
                NotImplemented)

    @abc.abstractmethod
    def initialize(self, model):
        """
        Create the results object and its elements based on capture_elemetns.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, model):
        """
        Update the results object at each iteration at which resutls are
        stored.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def postprocess(self, **kwargs):
        """
        Perform different tasks at the time of returning the results object.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def add_run_elements(self, model):
        """
        Add elements with run cache to the results object.
        """
        raise NotImplementedError


class DatasetHandler(OutputHandlerInterface):
    """
    Manages simulation results stored as netCDF4 Dataset.
    """

    def __init__(self, out_file):
        self.out_file = out_file
        self.ds = None
        self.__step = 0
        self.nc = __import__("netCDF4")

    def initialize(self, model):
        """
        Creates a netCDF4 Dataset and adds model dimensions and
        variables present in the capture elements to it.

        Parameters
        ----------
        model: pysd.Model
            PySD Model object

        Returns
        -------
        None

        """
        self.__step = 0
        self.ds = self.nc.Dataset(self.out_file, "w")

        # defining global attributes
        self.ds.description = "Results for simulation run on " \
            f"{t.ctime(t.time())} using PySD version {__version__}"
        self.ds.model_file = model.py_model_file or model.mdl_file
        self.ds.timestep = f"{model.time.time_step()}" if model.cache_type[
            "time_step"] == "run" else "Variable"
        self.ds.initial_time = f"{model.time.initial_time()}"
        self.ds.final_time = f"{model.time.final_time()}" if model.cache_type[
            "final_time"] == "run" else "Variable"

        # creating variables for all model dimensions
        for dim_name, coords in model.subscripts.items():
            coords = np.array(coords)
            # create dimension
            self.ds.createDimension(dim_name, len(coords))
            # length of the longest string in the coords
            max_str_len = len(max(coords, key=len))
            # create variable for the dimension
            var = self.ds.createVariable(
                dim_name, f"S{max_str_len}", (dim_name,))
            # assigning coords to dimension
            var[:] = coords

        # creating the time dimension as unlimited
        self.ds.createDimension("time", None)
        # creating variables
        self.__create_ds_vars(model, self.capture_elements_step + ['time'])

    def update(self, model):
        """
        Writes values of cache step variables from the capture_elements
        list in the netCDF4 Dataset.

        Parameters
        ----------
        model: pysd.Model
            PySD Model object

        Returns
        -------
        None

        """
        self.ds['time'][self.__step] = model.time.round()
        for key in self.capture_elements_step:
            comp = model[key]
            if isinstance(comp, xr.DataArray):
                self.ds[key][self.__step, :] = comp.values
            else:
                self.ds[key][self.__step] = comp

        self.__step += 1

    def __update_run_elements(self, model):
        """
        Writes values of cache run elements from the capture_elements
        set in the netCDF4 Dataset.
        Cache run elements do not have the time dimension.

        Parameters
        ----------
        model: pysd.Model
            PySD Model object

        Returns
        -------
        None

        """
        for key in self.capture_elements_run:
            comp = model[key]
            if isinstance(comp, xr.DataArray):
                self.ds[key][:] = comp.values
            else:
                self.ds[key][:] = comp

    def postprocess(self, **kwargs):
        """
        Closes netCDF4 Dataset.

        Returns
        -------
        None

        """
        self.ds.close()
        print(f"Results stored in {self.out_file}")

    def add_run_elements(self, model):
        """
        Adds constant elements to netCDF4 Dataset.

        Parameters
        ----------
        model: pysd.Model
            PySD Model object

        Returns
        -------
        None

        """
        # creating variables in capture_elements
        self.__create_ds_vars(model, self.capture_elements_run, time_dim=False)
        self.__update_run_elements(model)

    def __create_ds_vars(self, model, capture_elements, time_dim=True):
        """
        Create new variables in a netCDF4 Dataset from the capture_elements.
        Data is zlib compressed by default for netCDF4 1.6.0 and above.

        Parameters
        ----------
        model: pysd.Model
            PySD Model object.
        capture_elements: list
            List of variable or parameter names to include as variables in the
            dataset.
        time_dim: bool
            Whether to add time as the first dimension for the variable.

        Returns
        -------
        None

        """
        kwargs = dict()

        if tuple(self.nc.__version__.split(".")) >= ('1', '6', '0'):
            kwargs["compression"] = "zlib"

        for key in capture_elements:
            comp = model[key]

            dims = tuple()
            if isinstance(comp, xr.DataArray):
                dims = tuple(comp.dims)
            if time_dim:
                dims = ("time",) + dims

            var = self.ds.createVariable(key, "f8", dims, **kwargs)
            # adding metadata for each var from the model.doc
            for col in model.doc.columns:
                if col in ["Subscripts", "Limits"]:
                    # pass those that cannot be saved as attributes
                    continue
                var.setncattr(
                    col,
                    model.doc.loc[model.doc["Py Name"] == key, col].values[0]
                    or "Missing"
                    )


class DataFrameHandler(OutputHandlerInterface):
    """
    Manages simulation results stored as pandas DataFrame.
    """
    def __init__(self, out_file):
        self.out_file = out_file
        self.ds = None
        self.__step = 0

    def initialize(self, model):
        """
        Creates an empty dictionary to save the outputs.

        Parameters
        ----------
        model: pysd.Model
            PySD Model object

        Returns
        -------
        None

        """
        self.ds = defaultdict(list)
        self.__step = 0

    def update(self, model):
        """
        Add new values to the data dictionary.

        Parameters
        ----------
        model: pysd.Model
            PySD Model object

        Returns
        -------
        None

        """
        self.ds['time'].append(model.time.round())
        for key in self.capture_elements_step:
            self.ds[key].append(getattr(model.components, key)())

        self.__step += 1

    def postprocess(self, **kwargs):
        """
        Convert the output dictionary to a pandas DataFrame and flatten
        xarrays if required.

        Returns
        -------
        ds: pandas.DataFrame
            Simulation results stored as a pandas DataFrame.

        """
        # create the dataframe
        df = pd.DataFrame.from_dict(self.ds)
        df.set_index('time', inplace=True)

        # enforce flattening if df is to be saved to csv or tab file
        flatten = True if self.out_file else kwargs.get("flatten", None)

        df = DataFrameHandler.make_flat_df(
            df, kwargs["return_addresses"], flatten
            )
        if self.out_file:
            NCFile.df_to_text_file(df, self.out_file)

        return df

    def add_run_elements(self, model):
        """
        Adds constant elements to the output data dictionary.

        Parameters
        ----------
        model: pysd.Model
            PySD Model object

        Returns
        -------
        None

        """
        for key in self.capture_elements_run:
            self.ds[key] = [getattr(model.components, key)()] * self.__step

    @staticmethod
    def make_flat_df(df, return_addresses, flatten=False):
        """
        Takes a dataframe from the outputs of the integration processes,
        renames the columns as the given return_adresses and splits
        xarrays if needed.

        Parameters
        ----------
        df: pandas.DataFrame
            Dataframe to process.

        return_addresses: dict
            Keys will be column names of the resulting dataframe, and are what
            the user passed in as 'return_columns'. Values are a tuple:
            (py_name, {coords dictionary}) which tells us where to look for the
            value to put in that specific column.

        flatten: bool (optional)
                If True, once the output dataframe has been formatted will
                split the xarrays in new columns following vensim's naming
                to make a totally flat output. Default is False.

        Returns
        -------
        new_df: pandas.DataFrame
            Formatted dataframe.

        """
        new_df = {}
        for real_name, (pyname, address) in return_addresses.items():
            if address:
                # subset the specific address
                values = [x.loc[address] for x in df[pyname].values]
            else:
                # get the full column
                values = df[pyname].to_list()

            is_dataarray = len(values) != 0 and isinstance(
                values[0], xr.DataArray)

            if is_dataarray and values[0].size == 1:
                # some elements are returned as 0-d arrays, convert
                # them to float
                values = [x.squeeze().values[()] for x in values]
                is_dataarray = False

            if flatten and is_dataarray:
                DataFrameHandler.__add_flat(new_df, real_name, values)
            else:
                new_df[real_name] = values

        return pd.DataFrame(index=df.index, data=new_df)

    @staticmethod
    def __add_flat(savedict, name, values):
        """
        Add float lists from a list of xarrays to a provided dictionary.

        Parameters
        ----------
        savedict: dict
            Dictionary to save the data on.

        name: str
            The base name of the variable to save the data.

        values: list
            List of xarrays to convert to split in floats.

        Returns
        -------
        None

        """
        # remove subscripts from name if given
        name = re.sub(r'\[.*\]', '', name)
        dims = values[0].dims

        # split values in xarray.DataArray
        lval = [xrsplit(val) for val in values]
        for i, ar in enumerate(lval[0]):
            vals = [float(v[i]) for v in lval]
            subs = '[' + ','.join([str(ar.coords[dim].values)
                                   for dim in dims]) + ']'
            savedict[name+subs] = vals


class ModelOutput():
    """
    Manages outputs from simulations. Handles different types of outputs
    by dispatchinging the tasks to adequate object handlers.

    Parameters
    ----------
    out_file: str or pathlib.Path
        Path to the file where the results will be written.

    """
    out_handlers = {
        "__default__": DataFrameHandler,
        ".csv": DataFrameHandler,
        ".tab": DataFrameHandler,
        ".nc": DatasetHandler,
    }

    def __init__(self, out_file=None):
        self.handler = ModelOutput.get_handler(out_file)

    @staticmethod
    def get_handler(out_file):
        if out_file is None:
            return ModelOutput.out_handlers["__default__"](None)

        out_file = Path(out_file)

        try:
            return ModelOutput.out_handlers[out_file.suffix](out_file)
        except KeyError:
            raise ValueError(
                f"Unsupported output file format {out_file.suffix}")

    def set_capture_elements(self, capture_elements):
        self.handler.capture_elements_step = capture_elements["step"]
        self.handler.capture_elements_run = capture_elements["run"]

    def initialize(self, model):
        """
        Delegating the creation of the results object and its elements
        to the appropriate handler.
        """
        self.handler.initialize(model)

    def update(self, model):
        """
        Delegating the update of the results object and its elements
        to the appropriate handler.
        """
        self.handler.update(model)

    def postprocess(self, **kwargs):
        """
        Delegating the postprocessing of the results object
        to the appropriate handler.
        """
        return self.handler.postprocess(**kwargs)

    def add_run_elements(self, model):
        """
        Delegating the addition of results with run cache in the
        output object to the appropriate handler.
        """
        self.handler.add_run_elements(model)

    @staticmethod
    def collect(model, flatten_output=True):
        """
        Collect results after one or more simulation steps, and save to
        desired output format (DataFrame, csv, tab or netCDF).

        Parameters
        ----------
        model: pysd.py_backend.model.Model
            PySD Model object.

        flatten_output: bool (optional)
            If True, once the output dataframe has been formatted will
            split the xarrays in new columns following Vensim's naming
            to make a totally flat output. Default is True.
            This argument will be ignored when passing a netCDF4 file
            path in the output_file argument.

        """
        del model._dependencies["OUTPUTS"]

        model.output.add_run_elements(model)

        model._remove_constant_cache()

        return model.output.postprocess(
            return_addresses=model.return_addresses, flatten=flatten_output)



"""
The AST structures are created with the help of the parsimonious visitors
using the structures dictionary.

"""
import re
from ..structures import abstract_expressions as ae


structures = {
    "reference":  ae.ReferenceStructure,
    "subscripts_ref": ae.SubscriptsReferenceStructure,
    "arithmetic": ae.ArithmeticStructure,
    "logic": ae.LogicStructure,
    "with_lookup": ae.InlineLookupsStructure,
    "call": ae.CallStructure,
    "game": ae.GameStructure,
    "allocate_available": ae.AllocateAvailableStructure,
    "allocate_by_priority": ae.AllocateByPriorityStructure,
    "get_xls_lookups": ae.GetLookupsStructure,
    "get_direct_lookups": ae.GetLookupsStructure,
    "get_xls_data": ae.GetDataStructure,
    "get_direct_data": ae.GetDataStructure,
    "get_xls_constants": ae.GetConstantsStructure,
    "get_direct_constants": ae.GetConstantsStructure,
    "initial": ae.InitialStructure,
    "integ": ae.IntegStructure,
    "delay1": lambda x, y: ae.DelayStructure(x, y, x, 1),
    "delay1i": lambda x, y, z: ae.DelayStructure(x, y, z, 1),
    "delay3": lambda x, y: ae.DelayStructure(x, y, x, 3),
    "delay3i": lambda x, y, z: ae.DelayStructure(x, y, z, 3),
    "delay_n": ae.DelayNStructure,
    "delay_fixed": ae.DelayFixedStructure,
    "smooth": lambda x, y: ae.SmoothStructure(x, y, x, 1),
    "smoothi": lambda x, y, z: ae.SmoothStructure(x, y, z, 1),
    "smooth3": lambda x, y: ae.SmoothStructure(x, y, x, 3),
    "smooth3i": lambda x, y, z: ae.SmoothStructure(x, y, z, 3),
    "smooth_n": ae.SmoothNStructure,
    "trend": ae.TrendStructure,
    "forecast": lambda x, y, z: ae.ForecastStructure(x, y, z, 0),
    "sample_if_true": ae.SampleIfTrueStructure,
    "lookup": ae.LookupsStructure,
    "data": ae.DataStructure,
    "pulse_train": lambda start, width, interval, end: ae.CallStructure(
            ae.ReferenceStructure("pulse_train"),
            (start, interval, width, end))
}

operators = {
    "logic_ops": [":AND:", ":OR:"],
    "not_ops": [":NOT:"],
    "comp_ops": ["=", "<>", "<=", "<", ">=", ">"],
    "add_ops": ["+", "-"],
    "prod_ops": ["*", "/"],
    "exp_ops": ["^"],
    "pre_ops": ["+", "-"]
}


parsing_ops = {
    key: "|".join(re.escape(x) for x in values)
    for key, values in operators.items()
}



import warnings
from pathlib import Path
import numpy as np
from typing import List

from pysd.translators.structures.abstract_model import AbstractSubscriptRange
from pysd.py_backend.external import ExtSubscript


class SubscriptManager:
    """
    SubscriptManager object allows saving the subscripts included in the
    Section, searching for elements or keys and simplifying them.

    Parameters
    ----------
    abstrac_subscripts: list
        List of the AbstractSubscriptRanges comming from the AbstractModel.

    _root: pathlib.Path
        Path to the model file. Needed to read subscript ranges from
        Excel files.

    """
    def __init__(self, abstract_subscripts: List[AbstractSubscriptRange],
                 _root: Path):
        self._root = _root
        self._copied = []
        self.mapping = {}
        self.subscripts = abstract_subscripts
        self.elements = {}
        self.subranges = self._get_main_subscripts()
        self.subscript2num = self._get_subscript2num()

    @property
    def subscripts(self) -> dict:
        return self._subscripts

    @subscripts.setter
    def subscripts(self, abstract_subscripts: List[AbstractSubscriptRange]):
        self._subscripts = {}
        missing = []
        for sub in abstract_subscripts:
            self.mapping[sub.name] = sub.mapping
            if isinstance(sub.subscripts, list):
                # regular definition of subscripts
                self._subscripts[sub.name] = sub.subscripts
            elif isinstance(sub.subscripts, str):
                # copied subscripts, this will be always a subrange,
                # then we need to prevent them of being saved as a main range
                self._copied.append(sub.name)
                self.mapping[sub.name].append(sub.subscripts)
                if sub.subscripts in self._subscripts:
                    self._subscripts[sub.name] =\
                        self._subscripts[sub.subscripts]
                else:
                    missing.append(sub)
            elif isinstance(sub.subscripts, dict):
                # subscript from file
                self._subscripts[sub.name] = ExtSubscript(
                    file_name=sub.subscripts["file"],
                    tab=sub.subscripts["tab"],
                    firstcell=sub.subscripts["firstcell"],
                    lastcell=sub.subscripts["lastcell"],
                    prefix=sub.subscripts["prefix"],
                    root=self._root).subscript

                if not self._subscripts[sub.name]:
                    raise ValueError(
                        f"Subscript range '{sub.name}' empty:\n\t"
                        + str(sub.subscripts))
            else:
                raise ValueError(
                    f"Invalid definition of subscript '{sub.name}':\n\t"
                    + str(sub.subscripts))

        while missing:
            # second loop for copied subscripts
            sub = missing.pop()
            self._subscripts[sub.name] =\
                self._subscripts[sub.subscripts]

        subs2visit = self.subscripts.keys()
        while subs2visit:
            # third loop for subscripts defined with subranges
            updated = []
            for dim in subs2visit:
                if any(sub in self._subscripts
                       for sub in self._subscripts[dim]):
                    # a subrange name is being used to define the range
                    # subscripts
                    updated.append(dim)
                    new_subs = []
                    for sub in self._subscripts[dim]:
                        if sub in self.subscripts:
                            # append the subscripts of the subrange
                            new_subs += self._subscripts[sub]
                        else:
                            # append the same subscript
                            new_subs.append(sub)
                    self._subscripts[dim] = new_subs
            # visit again the updated ranges as there could be several
            # levels of subranges
            subs2visit = updated.copy()

    def _get_main_subscripts(self) -> dict:
        """
        Reutrns a dictionary with the main ranges as keys and their
        subranges as values.
        """
        subscript_sets = {
            name: set(subs) for name, subs in self.subscripts.items()}

        subranges = {}
        for range, subs in subscript_sets.items():
            # current subscript range
            subranges[range] = []
            for subrange, subs2 in subscript_sets.items():
                if range == subrange:
                    # pass current range
                    continue
                elif subs == subs2:
                    # range is equal to the subrange, as Vensim does
                    # the main range will be the first one alphabetically
                    # make it case insensitive
                    range_l = range.replace(" ", "_").lower()
                    subrange_l = subrange.replace(" ", "_").lower()
                    if range_l < subrange_l and range not in self._copied:
                        subranges[range].append(subrange)
                    else:
                        # copied subscripts ranges or subscripts ranges
                        # that come later alphabetically
                        del subranges[range]
                        break
                elif subs2.issubset(subs):
                    # subrange is a subset of range, append it to the list
                    subranges[range].append(subrange)
                elif subs2.issuperset(subs):
                    # it exist a range that contents the elements of the range
                    del subranges[range]
                    break

        return subranges

    def _get_subscript2num(self) -> dict:
        """
        Build a dictionary to return the numeric value or values of a
        subscript or subscript range.
        """
        s2n = {}
        for range, subranges in self.subranges.items():
            # a main range is direct to return
            s2n[range.replace(" ", "_").lower()] = (
                f"np.arange(1, len(_subscript_dict['{range}'])+1)",
                {range: self.subscripts[range]}
            )
            for i, sub in enumerate(self.subscripts[range], start=1):
                # a subscript must return its numeric position
                # in the main range
                s2n[sub.replace(" ", "_").lower()] = (str(i), {})
            for subrange in subranges:
                # subranges may return the position of each subscript
                # in the main range
                sub_index = [
                    self.subscripts[range].index(sub)+1
                    for sub in self.subscripts[subrange]]

                if np.all(
                  sub_index
                  == np.arange(sub_index[0], sub_index[0]+len(sub_index))):
                    # subrange definition can be simplified with a range
                    subsarray = f"np.arange({sub_index[0]}, "\
                        f"len(_subscript_dict['{subrange}'])+{sub_index[0]})"
                else:
                    # subrange definition cannot be simplified
                    subsarray = f"np.array({sub_index})"

                s2n[subrange.replace(" ", "_").lower()] = (
                    subsarray,
                    {subrange: self.subscripts[subrange]}
                )

        return s2n

    def _find_subscript_name(self, element: str, avoid: List[str] = []) -> str:
        """
        Given a member of a subscript family, return the first key of
        which the member is within the value list.

        Parameters
        ----------
        element: str
            Subscript or subscriptrange name to find.
        avoid: list (optional)
            List of subscripts to avoid. Default is an empty list.

        Returns
        -------
        name: str
            The first key of which the member is within the value list
            in the subscripts dictionary.

        Examples
        --------
        >>> sm = SubscriptManager([], Path(''))
        >>> sm._subscripts = {
        ...     'Dim1': ['A', 'B', 'C'],
        ...     'Dim2': ['A', 'B', 'C', 'D']}
        >>> sm._find_subscript_name('D')
        'Dim2'
        >>> sm._find_subscript_name('B')
        'Dim1'
        >>> sm._find_subscript_name('B', avoid=['Dim1'])
        'Dim2'

        """
        for name, elements in self.subscripts.items():
            if element in elements and name not in avoid:
                return name

    def make_coord_dict(self, subs: List[str]) -> dict:
        """
        This is for assisting with the lookup of a particular element.

        Parameters
        ----------
        subs: list of strings
            Coordinates, either as names of dimensions, or positions within
            a dimension.

        Returns
        -------
        coordinates: dict
            Coordinates needed to access the xarray quantities we are
            interested in.

        Examples
        --------
        >>> sm = SubscriptManager([], Path(''))
        >>> sm._subscripts = {
        ...     'Dim1': ['A', 'B', 'C'],
        ...     'Dim2': ['A', 'B', 'C', 'D']}
        >>> sm.make_coord_dict(['Dim1', 'D'])
        {'Dim1': ['A', 'B', 'C'], 'Dim2': ['D']}
        >>> sm.make_coord_dict(['A'])
        {'Dim1': ['A']}
        >>> sm.make_coord_dict(['A', 'B'])
        {'Dim1': ['A'], 'Dim2': ['B']}
        >>> sm.make_coord_dict(['A', 'Dim1'])
        {'Dim2': ['A'], 'Dim1': ['A', 'B', 'C']}

        """
        sub_elems_list = [y for x in self.subscripts.values() for y in x]
        coordinates = {}
        for sub in subs:
            if sub in sub_elems_list:
                name = self._find_subscript_name(
                    sub, avoid=subs + list(coordinates))
                coordinates[name] = [sub]
            else:
                if sub.endswith("!"):
                    coordinates[sub] = self.subscripts[sub[:-1]]
                else:
                    coordinates[sub] = self.subscripts[sub]
        return coordinates

    def make_merge_list(self, subs_list: List[List[str]],
                        element: str = "") -> List[str]:
        """
        This is for assisting when building xrmerge. From a list of subscript
        lists returns the final subscript list after merging. Necessary when
        merging variables with subscripts comming from different definitions.

        Parameters
        ----------
        subs_list: list of lists of strings
            Coordinates, either as names of dimensions, or positions within
            a dimension.
        element: str (optional)
            Element name, if given it will be printed with any error or
            warning message. Default is "".

        Returns
        -------
        dims: list
            Final subscripts after merging.

        Examples
        --------
        >>> sm = SubscriptManager([], Path(''))
        >>> sm._subscripts = {"upper": ["A", "B"], "all": ["A", "B", "C"]}
        >>> sm.make_merge_list([['A'], ['B']])
        ['upper']
        >>> sm.make_merge_list([['A'], ['B'], ['C']])
        ['all']
        >>> sm.make_merge_list([['upper'], ['C']])
        ['all']
        >>> sm.make_merge_list([['A'], ['C']])
        ['all']

        """
        def _to_list(dim):
            try:
                # convert array of dimension to a list
                return dim.tolist()
            except AttributeError:
                # few specific cases input is a string
                return dim

        coords_set = [set() for i in range(len(subs_list[0]))]
        coords_list = [
            self.make_coord_dict(subs)
            for subs in subs_list
        ]

        # update coords set
        [[coords_set[i].update(coords[dim]) for i, dim in enumerate(coords)]
         for coords in coords_list]

        dims = [None] * len(coords_set)
        # create an array with the name of the subranges for all
        # merging elements
        dims_list = np.array([
            list(coords) for coords in coords_list]).transpose()
        indexes = np.arange(len(dims))

        for i, coord2 in enumerate(coords_set):
            dims1 = [
                _to_list(dim) for dim in dims_list[i]
                if dim is not None and set(self.subscripts[dim]) == coord2
            ]
            if dims1:
                # if the given coordinate already matches return it
                dims[i] = dims1[0]
            else:
                # find a suitable coordinate
                other_dims = dims_list[indexes != i]
                for name, elements in self.subscripts.items():
                    if coord2 == set(elements) and name not in other_dims:
                        dims[i] = name
                        break

                if not dims[i]:
                    # the dimension is incomplete use the smaller
                    # dimension that completes it
                    for name, elements in self.subscripts.items():
                        if coord2.issubset(set(elements))\
                           and name not in other_dims:
                            dims[i] = name
                            warnings.warn(
                                element
                                + "\nDimension given by subscripts:"
                                + "\n\t{}\nis incomplete ".format(coord2)
                                + "using {} instead.".format(name)
                                + "\nSubscript_dict:"
                                + "\n\t{}".format(self.subscripts)
                            )
                            break

                if not dims[i]:
                    for name, elements in self.subscripts.items():
                        if coord2 == set(elements):
                            j = 1
                            while name + str(j) in self.subscripts.keys():
                                j += 1
                            self.subscripts[name + str(j)] = elements
                            dims[i] = name + str(j)
                            warnings.warn(
                                element
                                + "\nAdding new subscript range to"
                                + " subscript_dict:\n"
                                + name + str(j) + ": " + ', '.join(elements))
                            break

        return dims

    def simplify_subscript_input(self, coords: dict,
                                 merge_subs: List[str] = None) -> tuple:
        """
        Simplifies the subscripts input to avoid printing the coordinates
        list when the _subscript_dict can be used. Makes model code more
        simple.

        Parameters
        ----------
        coords: dict
            Coordinates to write in the model file.

        merge_subs: list of strings or None (optional)
            List of the final subscript range of the Python array after
            merging with other objects. If None the merge_subs will be
            taken from coords. Default is None.

        Returns
        -------
        final_subs, coords: dict, str
            Final subscripts and the equations to generate the coord
            dicttionary in the model file.

        Examples
        --------
        >>> sm = SubscriptManager([], Path(''))
        >>> sm._subscripts = {
        ...     "dim": ["A", "B", "C"],
        ...     "dim2": ["A", "B", "C", "D"]}
        >>> sm.simplify_subscript_input({"dim": ["A", "B", "C"]})
        ({"dim": ["A", "B", "C"]}, "{'dim': _subscript_dict['dim']}"
        >>> sm.simplify_subscript_input({"dim": ["A", "B", "C"]}, ["dim2"])
        ({"dim2": ["A", "B", "C"]}, "{'dim2': _subscript_dict['dim']}"
        >>> sm.simplify_subscript_input({"dim": ["A", "B"]})
        ({"dim": ["A", "B"]}, "{'dim': ['A', 'B']}"

        """
        if merge_subs is None:
            merge_subs = list(coords)

        coordsp = []
        final_subs = {}
        for ndim, (dim, coord) in zip(merge_subs, coords.items()):
            # find dimensions can be retrieved from _subscript_dict
            final_subs[ndim] = coord
            if not dim.endswith("!") and coord == self.subscripts[dim]:
                # use _subscript_dict
                coordsp.append(f"'{ndim}': _subscript_dict['{dim}']")
            else:
                # write whole dict
                coordsp.append(f"'{ndim}': {coord}")

        return final_subs, "{" + ", ".join(coordsp) + "}"



import warnings

import pandas as pd
import numpy as np
import xarray as xr

from . import utils


class Lookups(object):
    def set_values(self, values):
        """Set new values from user input"""
        self.data = xr.DataArray(
            np.nan, self.final_coords, list(self.final_coords))

        if isinstance(values, pd.Series):
            index = list(values.index)
            index.sort()
            self.data = self.data.expand_dims(
                {'lookup_dim': index}, axis=0).copy()

            for index, value in values.items():
                if isinstance(values.values[0], xr.DataArray):
                    self.data.loc[index].loc[value.coords] = value
                else:
                    self.data.loc[index] = value
        else:
            if isinstance(values, xr.DataArray):
                self.data.loc[values.coords] = values.values
            else:
                if self.final_coords:
                    self.data.loc[:] = values
                else:
                    self.data = values

    def __call__(self, x, final_subs=None):
        try:
            return self._call(self.data, x, final_subs)
        except (TypeError, KeyError):
            # this except catch the errors when a lookups has been
            # changed to a constant value by the user
            if final_subs and isinstance(self.data, xr.DataArray):
                # self.data is an array, reshape it
                outdata = xr.DataArray(np.nan, final_subs, list(final_subs))
                return xr.broadcast(outdata, self.data)[1]
            elif final_subs:
                # self.data is a float, create an array
                return xr.DataArray(self.data, final_subs, list(final_subs))
            else:
                return self.data

    def _call(self, data, x, final_subs=None):
        if isinstance(x, xr.DataArray):
            if not x.dims:
                # shape 0 xarrays
                return self._call(data, float(x))

            outdata = xr.DataArray(np.nan, final_subs, list(final_subs))

            if self.interp != "extrapolate" and\
               np.all(x > data['lookup_dim'].values[-1]):
                outdata_ext = data[-1]
                warnings.warn(
                  self.py_name + "\n"
                  + "extrapolating data above the maximum value of the series")
            elif self.interp != "extrapolate" and\
              np.all(x < data['lookup_dim'].values[0]):
                outdata_ext = data[0]
                warnings.warn(
                  self.py_name + "\n"
                  + "extrapolating data below the minimum value of the series")
            else:
                data = xr.broadcast(data, x)[0]
                for a in utils.xrsplit(x):
                    outdata.loc[a.coords] = self._call(
                        data.loc[a.coords],
                        float(a))
                return outdata

            # return the final array in the specified dimensions order
            return xr.broadcast(
                outdata, outdata_ext.reset_coords('lookup_dim', drop=True))[1]

        else:
            if x in data['lookup_dim'].values:
                outdata = data.sel(lookup_dim=x)
            elif x > data['lookup_dim'].values[-1]:
                if self.interp == "extrapolate":
                    # extrapolate method for xmile models
                    k = (data[-1]-data[-2])\
                        / (data['lookup_dim'].values[-1]
                           - data['lookup_dim'].values[-2])
                    outdata = data[-1] + k*(x - data['lookup_dim'].values[-1])
                else:
                    outdata = data[-1]
                warnings.warn(
                  self.py_name + "\n"
                  + "extrapolating data above the maximum value of the series")
            elif x < data['lookup_dim'].values[0]:
                if self.interp == "extrapolate":
                    # extrapolate method for xmile models
                    k = (data[1]-data[0])\
                        / (data['lookup_dim'].values[1]
                           - data['lookup_dim'].values[0])
                    outdata = data[0] + k*(x - data['lookup_dim'].values[0])
                else:
                    outdata = data[0]
                warnings.warn(
                  self.py_name + "\n"
                  + "extrapolating data below the minimum value of the series")
            elif self.interp == 'hold_backward':
                outdata = data.sel(lookup_dim=x, method="pad")
            else:
                outdata = data.interp(lookup_dim=x)

            # the output could be a float or an xarray
            if self.is_float:
                # if lookup has no-coords return a float
                return float(outdata)
            else:
                # Remove lookup dimension coord from the DataArray
                return outdata.reset_coords('lookup_dim', drop=True)


class HardcodedLookups(Lookups):
    """Class for lookups defined in the file"""

    def __init__(self, x, y, coords, interp, final_coords, py_name):
        # TODO: avoid add and merge all declarations in one definition
        self.is_float = not bool(coords)
        self.py_name = py_name
        self.final_coords = final_coords
        self.values = [(x, y, coords)]
        self.interp = interp

    def add(self, x, y, coords):
        self.values.append((x, y, coords))

    def initialize(self):
        """
        Initialize all elements and create the self.data xarray.DataArray
        """
        if len(self.values) == 1:
            # Just loag one value (no add)
            for x, y, coords in self.values:
                if len(x) != len(set(x)):
                    raise ValueError(
                        self.py_name + "\n"
                        "x dimension has repeated values..."
                    )
                try:
                    y = np.array(y).reshape((len(x),) + (1,)*len(coords))
                except ValueError:
                    raise ValueError(
                        self.py_name + "\n"
                        "x and y dimensions have different length..."
                    )
                self.data = xr.DataArray(
                    np.tile(y, [1] + utils.compute_shape(coords)),
                    {"lookup_dim": x, **coords},
                    ["lookup_dim"] + list(coords)
                )
        else:
            # Load in several lines (add)
            self.data = xr.DataArray(
                np.nan, self.final_coords, list(self.final_coords))

            for x, y, coords in self.values:
                if "lookup_dim" not in self.data.dims:
                    # include lookup_dim dimension in the final array
                    self.data = self.data.expand_dims(
                        {"lookup_dim": x}, axis=0).copy()
                else:
                    # add new coordinates (if needed) to lookup_dim
                    x_old = list(self.data.lookup_dim.values)
                    x_new = list(set(x).difference(x_old))
                    self.data = self.data.reindex(lookup_dim=x_old+x_new)

                # reshape y value and assign it to self.data
                y = np.array(y).reshape((len(x),) + (1,)*len(coords))
                self.data.loc[{"lookup_dim": x, **coords}] =\
                    np.tile(y, [1] + utils.compute_shape(coords))

        # sort data
        self.data = self.data.sortby("lookup_dim")

        if np.any(np.isnan(self.data)):
            # fill missing values of different input lookup_dim values
            values = self.data.values
            self._fill_missing(self.data.lookup_dim.values, values)
            self.data = xr.DataArray(values, self.data.coords, self.data.dims)

    def _fill_missing(self, series,  data):
        """
        Fills missing values in lookups to have a common series.
        Mutates the values in data.

        Returns
        -------
        None

        """
        if len(data.shape) > 1:
            # break the data array until arrive to a vector
            for i in range(data.shape[1]):
                if np.any(np.isnan(data[:, i])):
                    self._fill_missing(series, data[:, i])
        elif not np.all(np.isnan(data)):
            # interpolate missing values
            data[np.isnan(data)] = np.interp(
                series[np.isnan(data)],
                series[~np.isnan(data)],
                data[~np.isnan(data)])



"""
The provided allocation functions have no direct analog in the standard
Python data analytics stack. They are provided in a structure that makes
it easy for the model elements to call. The functions may be similar to
the original functions given by Vensim, but sometimes the number or
order of arguments may change. The allocation functions may call a
protected function or class method thatintegrates the algorithm to
compute the allocation. The algorithms are briefly explained in these
functions docstring.

Note
----
The Allocation functions basis is explained in the Vensim documentation.
https://www.vensim.com/documentation/allocation_overview.html

Warning
-------
Some allocation function's results may differ from the result given by
Vensim as optimization functions are used to solve the allocation
problems. Those algorithms may not work in the same way or may
have differences in the numerical error propagation.

"""
import itertools
from math import erfc

import numpy as np
import xarray as xr
from scipy.optimize import least_squares
import portion as p


class Priorities:
    @classmethod
    def get_functions(cls, q0, pp, kind):
        """
        Get priority functions based on the demand/supply and priority profile.

        Parameters
        ----------
        q0: numpy.array
            values of maximum demand or supply of each component.
            Its shape should be (n,)
        pp: numpy.array
            pp values array. Its shape should be (n, m).
        kind: str ("demand" or "supply")
            The kind of priority "demand" or "supply".

        Returns
        -------
        functions: list of functions
            List of allocation supply or demand function for each element.

        full_allocation: function
            Full allocation function. It is the result function of
            addying all the functions.

        def_intervals: list of tuples
            List of (supply interval, priority interval, mean priority)
            where the full_allocation function is extrictly monotonous
            (injective). Givin a supply value, this is used to compute
            the limits and starting point of the optimization problem.

        """
        if np.any(pp[:, 2] <= 0):
            # pwidth values smaller than 0
            raise ValueError("pwidth values must be positive.")

        if kind == "demand":
            # Get the list of priority functions and the intervals where
            # they are strictly monotonous (injective function)
            func_int = [
                cls.get_function_demand(q0[i], pp[i])
                for i in range(pp.shape[0])
            ]
            # In order to get the range of the full_allocation function,
            # we need to flip the lower and the upper value, as it is a
            # decreasing function for demand
            int_attr = {"lower": "upper", "upper": "lower"}
        elif kind == "supply":  # pragma: no cover
            # Get the list of priority functions and the intervals where
            # they are strictly monotonous (injective function)
            func_int = [
                cls.get_function_supply(q0[i], pp[i])
                for i in range(pp.shape[0])
            ]
            # In order to get the range of the full_allocation function,
            # we need to keep the lower and the upper value, as it is a
            # increasing function for supply
            int_attr = {"lower": "lower", "upper": "upper"}
        else:
            raise ValueError(
                f"kind='{kind}' is not allowed. kind should be "
                "'demand' or 'supply'.")

        functions = [fi[0] for fi in func_int]
        intervals = [fi[1] for fi in func_int]

        # Join the intervals of all functions to get the intervals where
        # the sum of the functions is strictly monotonous (injective
        # function), therefore we can solve the minimization problem in
        # strictly monotonous areas of the function, avoiding the crash
        # of the algorithm
        interval = intervals[0]
        for i in intervals[1:]:
            interval = interval.union(i)

        # Full allocation function -> function to solve
        def full_allocation(x):
            if isinstance(x, np.ndarray):
                # Fix to solve issues in the newest numpy versions
                x = x.squeeze()[()]
            return np.sum([func(x) for func in functions])

        def_intervals = []
        for subinterval in interval:
            # Iterate over disjoint interval sections
            # Each interval section will be converted in supply interval
            # and compute the starting point for the supply interval
            # as the midpoint in the priority interval
            def_intervals.append((
                p.closed(
                    full_allocation(getattr(subinterval, int_attr["lower"])),
                    full_allocation(getattr(subinterval, int_attr["upper"]))
                ),
                subinterval,
                .5*(subinterval.upper+subinterval.lower)
            ))

        return functions, full_allocation, def_intervals

    @classmethod
    def get_function_demand(cls, q0, pp):
        """
        Get priority functions for demand based on the priority profile.

        Parameters
        ----------
        q0: float [0, +np.inf)
            The demand of the target.
        pp: numpy.array
            pp values array.

        Returns
        -------
        priority_func: function
            Priority function.
        interval: portion.interval
            The interval where the priority function is strictly monotonous.

        """
        if q0 == 0:
            # No demand is requested return a 0 function with an empty interval
            return lambda x: 0, p.empty()
        if pp[0] == 0:
            # Fixed quantity demand
            return cls.fixed_quantity(q0, *pp[1:])
        elif pp[0] == 1:
            # Rectangular demand
            return cls.rectangular(q0, *pp[1:])
        elif pp[0] == 2:
            # Triangular demand
            return cls.triangular(q0, *pp[1:])
        elif pp[0] == 3:
            # Normal distribution demand
            return cls.normal(q0, *pp[1:])
        elif pp[0] == 4:
            # Exponential distribution demand
            return cls.exponential(q0, *pp[1:])
        elif pp[0] == 5:
            # Constant elasticity demand
            return cls.constant_elasticity_demand(q0, *pp[1:])
        else:
            raise ValueError(
                f"The priority function for pprofile={pp[0]} is not valid.")

    @classmethod
    def get_function_supply(cls, q0, pp):
        """
        Get priority functions for supply based on the priority profile.

        Parameters
        ----------
        q0: float [0, +np.inf)
            The supply of the producer.
        pp: numpy.array
            pp values array.

        Returns
        -------
        priority_func: function
            Priority function.
        interval: portion.interval
            The interval where the priority function is strictly monotonous.

        """
        # TODO: This function should be similar to the demand function
        # it is neccessary for the many-to-many allocation given by
        # the set FIND MARKET PLACE, DEMAND AT PRICE, SUPPLY AT PRICE
        raise NotImplementedError("get_function_supply is not implemented.")

    @staticmethod
    def fixed_quantity(q0, ppriority, pwidth, pextra):
        raise NotImplementedError(
            "fixed_quantity priority profile is not implemented.")

    @staticmethod
    def rectangular(q0, ppriority, pwidth, pextra):
        """
        Demand curve for rectangular shape.
        The supply curve will be shaped as the integral of a rectangle.

        Parameters
        ----------
        q0: float
            The total demand/supply of the element.
        ppriority: float
            Specifies the midpoint of the curve.
        pwidth: float
            Determines the speed with which the curve goes from 0 to
            the specified quantity.
        pextra: float
            Ignore.

        Returns
        -------
        priority_func: function
            The priority function.

        """
        def priority_func(x):
            if x <= ppriority - pwidth*.5:
                return q0
            elif x < ppriority + pwidth*.5:
                return q0*(1-(x-ppriority+pwidth*.5)/pwidth)
            else:
                return 0

        return (
            priority_func,
            p.open(ppriority - pwidth*.5, ppriority + pwidth*.5)
        )

    @staticmethod
    def triangular(q0, ppriority, pwidth, pextra):
        """
        Demand curve for triangular shape.
        The supply curve will be shaped as the integral of a triangle.

        Parameters
        ----------
        q0: float
            The total demand/supply of the element.
        ppriority: float
            Specifies the midpoint of the curve.
        pwidth: float
            Determines the speed with which the curve goes from 0 to the
            specified quantity.
        pextra: float
            Ignore.

        Returns
        -------
        priority_func: function
            The priority function.

        """
        def priority_func(x):
            if x <= ppriority - pwidth*.5:
                return q0
            elif x < ppriority:
                return q0*(1-2*(x-ppriority+pwidth*.5)**2/pwidth**2)
            elif x < ppriority + pwidth*.5:
                return 2*q0*(ppriority+pwidth*.5-x)**2/pwidth**2
            else:
                return 0

        return (
            priority_func,
            p.open(ppriority - pwidth*.5, ppriority + pwidth*.5)
        )

    @staticmethod
    def normal(q0, ppriority, pwidth, pextra):
        """
        Demand curve for normal shape.
        The supply curve will be shaped as the integral of a normal
        distribution.

        Parameters
        ----------
        q0: float
            The total demand/supply of the element.
        ppriority: float
            Specifies the midpoint of the curve (the mean of the
            underlying distribution).
        pwidth: float
            Standard deviation of the underlying distribution.
        pextra: float
            Ignore.

        Returns
        -------
        priority_func: function
            The priority function.

        """
        def priority_func(x):
            return q0*.5*(2-erfc((ppriority-x)/(np.sqrt(2)*pwidth)))

        # Normal distribution CDF is stricty monotonous in (-inf, inf).
        # However, numerically it is only in a the range ~ (-8.29*sd, 8.29*sd)
        return (
            priority_func,
            p.open(
                ppriority-8.2923611*pwidth,
                ppriority+8.2923611*pwidth
            )
        )

    @staticmethod
    def exponential(q0, ppriority, pwidth, pextra):
        """
        Demand curve for exponential shape
        The supply curve will be shaped as the integral of an
        exponential distribution that is symmetric around its mean
        (0.5*exp(-ABS(x-ppriority)/pwidth) on -∞ to ∞).

        Parameters
        ----------
        q0: float
            The total demand/supply of the element.
        ppriority: float
            Specifies the midpoint of the curve (the mean of the
            underlying distribution).
        pwidth: float
            Multiplier on x in the underlying distribution.
        pextra: float
            Ignore.

        Returns
        -------
        priority_func: function
            The priority function.

        """
        def priority_func(x):
            if x < ppriority:
                return q0*(1-.5*np.exp((x-ppriority)/pwidth))
            else:
                return q0*.5*np.exp((ppriority-x)/pwidth)

        # Exponential distribution CDF is stricty monotonous in (-inf, inf).
        # However, numerically it is only in a the range ~ (-36.7*sd, 36.7*sd)
        return (
            priority_func,
            p.open(
                ppriority-36.7368005696*pwidth,
                ppriority+36.7368005696*pwidth
            )
        )

    @staticmethod
    def constant_elasticity_demand(q0, ppriority, pwidth, pextra):
        """
        Demand constant elasticity curve.
        The curve will be a constant elasticity curve.

        Parameters
        ----------
        q0: float
            The total demand/supply of the element.
        ppriority: float
            Specifies the midpoint of the curve (the mean of the
            underlying distribution).
        pwidth: float
            Standard deviation of the underlying distribution.
        pextra: positive float
            Elasticity exponent.

        Returns
        -------
        priority_func: function
            The priority function.

        """
        raise NotImplementedError(
            "Some results for Vensim showed some bugs when using this "
            "priority curve. Therefore, the curve is not implemented in "
            "PySD as it cannot be properly tested."
        )

    @staticmethod
    def constant_elasticity_supply(ppriority, pwidth,
                                   pextra):   # pragma: no cover
        """
        Supply constant elasticity curve.
        The curve will be a constant elasticity curve.

        Parameters
        ----------
        q0: float
            The total demand/supply of the element.
        ppriority: float
            Specifies the midpoint of the curve (the mean of the
            underlying distribution).
        pwidth: float
            Standard deviation of the underlying distribution.
        pextra: positive float
            Elasticity exponent.

        Returns
        -------
        priority_func: function
            The priority function.

        """
        raise NotImplementedError(
            "Some results for Vensim showed some bugs when using this "
            "priority curve. Therefore, the curve is not implemented in "
            "PySD as it cannot be properly tested."
        )


def _allocate_available_1d(request, pp, avail):
    """
    This function implements the algorithm for allocate_available
    to be passed for 1d numpy.arrays. The algorithm works as follows:

    0. If supply > sum(request): return request. In the same way,
       if supply = 0: return request*0
    1. Based on the priority profiles and demands, the priority profiles
       are computed. This profiles are returned with the interval where
       each of them is strictly monotonous (or injective).
    2. Using the intervals of injectivity the initial guess is
       selected depending on the available supply.
    3. The initial guess and injectivity interval are used to compute
       the value where the sum of all priority functions is equal to
       the avilable supply. This porcess is done using a least_squares
       optimization function.
    4. The output from the previous step is used to compute the supply
       to each target.

    Parameters
    ----------
    request: numpy.ndarray (1D)
        The request by target. Values must be non-negative.
    pp: numpy.ndarray (2D)
        The priority profiles of each target.
    avail: float
        The available supply. Must be non-negative.

    Returns
    -------
    out: numpy.ndarray (1D)
        The distribution of the supply.

    """
    if avail >= np.sum(request):
        return request
    if avail == 0:
        return np.zeros_like(request)

    priorities, full_allocation, intervals =\
        Priorities.get_functions(request, pp, "demand")

    for interval, x_interval, x0 in intervals:
        if avail in interval:
            break
    priority = least_squares(
        lambda x: full_allocation(x) - avail,
        x0,
        bounds=(x_interval.lower, x_interval.upper),
        method='dogbox',
        tr_solver='exact',
        ).x[0]

    return [allocate(priority) for allocate in priorities]


def allocate_available(request, pp, avail):
    """
    Implements Vensim's ALLOCATE AVAILABLE function.
    https://www.vensim.com/documentation/fn_allocate_available.html

    Parameters
    -----------
    request: xarray.DataArray
        Request of each target. Its shape should be the one of the
        expected output of the function, having the allocation dimension
        in the last position.
    pp: xarray.DataArray
        Priority of each target. Its shape should be the same as
        request with an extra dimension for the priority profiles
        in the last position. See Vensim's documentation for more
        information https://www.vensim.com/documentation/24335.html
    avail: float or xarray.DataArray
        The total supply available to fulfill all requests. If the
        supply exceeds total requests, all requests are filled, but
        none are overfilled. If you wish to conserve material you must
        compute supply minus total allocations explicitly. Its shape,
        should be the same of request without the last dimension.

    Returns
    -------
    out: xarray.DataArray
        The distribution of the supply.

    Warning
    -------
    This function uses an optimization method for resolution and the
    given solution could differ from the one from Vensim. Particularly,
    when close to the boundaries of the defined priority profiles.

    """
    if np.any(request < 0):
        raise ValueError(
            "There are some negative request values. Ensure that "
            "your request is always non-negative. Allocation requires "
            f"all quantities to be positive or 0.\n{request}")

    if np.any(avail < 0):
        raise ValueError(
            f"avail={avail} is not allowed. avail should be non-negative."
        )

    if len(request.shape) == 1:
        # NUMPY: avoid '.values' and return directly the result of the
        # function call
        return xr.DataArray(
            _allocate_available_1d(
                request.values, pp.values, avail),
            request.coords
        )

    # NUMPY: use np.empty_like and remove '.values'
    out = xr.zeros_like(request, dtype=float)
    for comb in itertools.product(*[range(i) for i in avail.shape]):
        out.values[comb] = _allocate_available_1d(
            request.values[comb], pp.values[comb], avail.values[comb])

    return out


def _allocate_by_priority_1d(request, priority, width, supply):
    """
    This function implements the algorithm for allocate_by_priority
    to be passed for 1d numpy.arrays. The algorithm works as follows:

    0. If supply > sum(request): return request.
    1. Order the request and priorities from bigger to lower priorities.
    2. Compute the 'distances' between the target, the 'distance' is
       defined as the difference between priorities divided by width
       multiplied by the target request. If the difference in priorities
       over width is bigger than 1, set it to 1, having the distance
       equal to the request. For example, priorities = [10, 9, 7, 2],
       request = [3, 6, 2.5, 2] and width = 3 will have the following
       'distances' vector, distance = [(10-9)/3*3, (9-7)/3*6, 1*2.5]
       = [1, 4, 2.5]
    3. The supply is assigned with linear functions. The fraction
       (or slope) of supply that a target receives is its total request
       divided by the request of all the targets that are receiving
       supply at that point.
    4. The supply is assigned from bigger to lower priority. Starts
       assigning the supply to the first target, when it reaches the
       quantity of the 'distance,' to the second target, the second
       target will start receiving its supply. When the second target
       receives its 'distance' to the third target, the third target
       will start receiving supply and so on.
    5. Each time a target reaches its request or a new target starts
       receiving supply the slope of each target is computed again.
    6. It finishes when all the supply is distributed between targets

    Parameters
    ----------
    request: numpy.ndarray (1D)
        The request by target. Values must be non-negative.
    priority: numpy.ndarray (1D)
        The priority of each target.
    width: float
        The width between priorities. Must be positive.
    supply: float
        The available supply. Must be non-negative.

    Returns
    -------
    out: numpy.ndarray (1D)
        The distribution of the supply.

    """
    if supply >= np.sum(request):
        # All targets receive their request
        return request
    elif supply == 0:
        # No supply, all targets receive 0
        return np.zeros_like(request)

    # Remove request 0 targets and order by priority
    is_0 = request == 0
    sort = (-priority[~is_0]).argsort()
    request = request[~is_0].astype(float)[sort]
    priority = priority[~is_0][sort]
    # Create the outputs array
    out_return = np.zeros_like(is_0, dtype=float)
    out = np.zeros_like(request, dtype=float)
    # Compute the distances between target supply and next target start
    distances = np.full_like(request, np.nan, dtype=float)
    # last target will have an numpy.nan as distances as there are no
    # more targets after
    distances[:-1] = np.minimum(-np.diff(priority)/width, 1)*request[:-1]
    # Create a vector of the current active targets
    active = np.zeros_like(request, dtype=bool)
    active[0] = True
    # Create a vector of the last activated target
    c_i = 0
    while supply > 0:
        # Compute the slopes of the active targets of supply
        slopes = request*active
        slopes /= np.sum(slopes)
        # Compute how much supply much be given to any target reach its request
        dx_next_top = np.nanmin((request-out)[active]/slopes[active])
        # Compute how much supply is needed to start next target
        # (last target will return a numpy.nan)
        dx_next_start = (distances[c_i]-out[c_i])/slopes[c_i]
        # Compute where the next change in allocation function will change
        # this will happen when a target reaches is request, or when
        # the next target starts or when the supply is totally distributed
        dx = np.nanmin((dx_next_top, dx_next_start, supply))
        # Assing the supply to the targets
        out += slopes*dx
        if np.isclose(dx, dx_next_start, rtol=1e-10, atol=1e-16):
            # A new target will start in the next loop
            c_i += 1
            # Active the next targetif its request is different than 0
            active[c_i] = True
        if dx == dx_next_top:
            # One or more target have reached their request
            active[out == request] = False
        supply -= dx
    # Return the distributed supply in the original order
    # adding to it again the request 0 if the where removed
    out_return[~is_0] = out[sort.argsort()]
    return out_return


def allocate_by_priority(request, priority, width, supply):
    """
    Implements Vensim's ALLOCATE BY PRIORITY function.
    https://www.vensim.com/documentation/fn_allocate_by_priority.html

    Parameters
    -----------
    request: xarray.DataArray
        Request of each target. Its shape should be the same as
        priority. width and supply must have the same shape except the
        last dimension.
    priority: xarray.DataArray
        Priority of each target. Its shape should be the same as
        request. width and supply must have the same shape except the
        last dimension.
    width: float or xarray.DataArray
        Specifies how big a gap in priority is required to have the
        allocation go first to higher priority with only leftovers going
        to lower priority. When the distance between any two priorities
        exceeds width and the higher priority does not receive its full
        request the lower priority will receive nothing. Its shape
        should be the same as supply.
    supply: float or xarray.DataArray
        The total supply available to fulfill all requests. If the
        supply exceeds total requests, all requests are filled, but
        none are overfilled.  If you wish to conserve material you must
        compute supply minus total allocations explicitly. Its shape
        should be the same as width.

    Returns
    -------
    out: xarray.DataArray
        The distribution of the supply.

    """
    if np.any(request < 0):
        raise ValueError(
            "There are some negative request values. Ensure that "
            "your request is always non-negative. Allocation requires "
            f"all quantities to be positive or 0.\n{request}")

    if np.any(width <= 0):
        raise ValueError(
            f"width={width} \n is not allowed. width must be greater than 0.")

    if np.any(supply < 0):
        raise ValueError(
            f"supply={supply} \n is not allowed. supply must not be negative.")

    if len(request.shape) == 1:
        # NUMPY: avoid '.values' and return directly the result of the
        # function call
        return xr.DataArray(
            _allocate_by_priority_1d(
                request.values, priority.values, width, supply),
            request.coords
        )

    # NUMPY: use np.empty_like and remove '.values'
    out = xr.zeros_like(request, dtype=float)
    for comb in itertools.product(*[range(i) for i in supply.shape]):
        out.values[comb] = _allocate_by_priority_1d(
            request.values[comb], priority.values[comb],
            width.values[comb], supply.values[comb])

    return out



import re

from unicodedata import normalize
from typing import List

# used to create Python safe names with the variable reserved_words
from keyword import kwlist
from builtins import __dir__ as bidir
from pysd.py_backend.allocation import __dir__ as adir
from pysd.py_backend.cache import __dir__ as cadir
from pysd.py_backend.components import __dir__ as cdir
from pysd.py_backend.data import __dir__ as ddir
from pysd.py_backend.external import __dir__ as edir
from pysd.py_backend.functions import __dir__ as fdir
from pysd.py_backend.statefuls import __dir__ as sdir
from pysd.py_backend.utils import __dir__ as udir


class NamespaceManager:
    """
    NamespaceManager object allows includying new elements to the namespace
    and searching for elements in the namespace. When includying new
    elements a Python safe name is used to be able to write the equations.

    Parameters
    ----------
    parameters: list (optional)
        List of the parameters that are used as argument in the Macro.
        By defaukt it is an empty list.

    """
    _reserved_words = set(
        dir() + adir() + bidir() + cadir() + cdir() + ddir() + edir()
        + fdir() + sdir() + udir()).union(kwlist)

    def __init__(self, parameters: List[str] = []):
        self._used_words = self._reserved_words.copy()
        # inlcude time to the namespace
        self.namespace = {"Time": "time"}
        # include time to the cleanspace (case and whitespace/underscore
        # insensitive namespace)
        self.cleanspace = {"time": "time"}
        for parameter in parameters:
            self.add_to_namespace(parameter)

    def add_to_namespace(self, string: str) -> None:
        """
        Add a new string to the namespace.

        Parameters
        ----------
        string: str
            String to add to the namespace.

        Returns
        -------
        None

        """
        self.make_python_identifier(string, add_to_namespace=True)

    def make_python_identifier(self, string: str, prefix: str = None,
                               add_to_namespace: bool = False) -> str:
        """
        Takes an arbitrary string and creates a valid Python identifier.

        If the Python identifier created is already in the namespace,
        but the input string is not (ie, two similar strings resolve to
        the same Python identifier) or if the identifier is a reserved
        word in the reserved_words list, or is a Python default
        reserved word, adds _1, or if _1 is in the namespace, _2, etc.

        Parameters
        ----------
        string: str
            The text to be converted into a valid Python identifier.

        prefix: str or None (optional)
            If given it will be used as a prefix for the output string.
            Default is None.

        add_to_namespace: bool (optional)
            If True it will add the passed string to the namespace and
            to the cleanspace. Default is False.

        Returns
        -------
        identifier: str
            A vaild Python identifier based on the input string.

        Examples
        --------
        >>> make_python_identifier('Capital')
        'capital'

        >>> make_python_identifier('multiple words')
        'multiple_words'

        >>> make_python_identifier('multiple     spaces')
        'multiple_spaces'

        When the name is a Python keyword, add '_1' to differentiate it
        >>> make_python_identifier('for')
        'for_1'

        Remove leading and trailing whitespace
        >>> make_python_identifier('  whitespace  ')
        'whitespace'

        Remove most special characters outright:
        >>> make_python_identifier('H@t tr!ck')
        'ht_trck'

        add valid string to leading digits
        >>> make_python_identifier('123abc')
        'nvs_123abc'

        already in namespace
        >>> make_python_identifier('Var$')  # namespace={'Var$': 'var'}
        'var'

        namespace conflicts
        >>> make_python_identifier('Var@')  # namespace={'Var$': 'var'}
        'var_1'

        >>> make_python_identifier('Var$')  # namespace={'Var@': 'var',
        ...                                              'Var%':'var_1'}
        'var_2'

        References
        ----------
        Identifiers must follow the convention outlined here:
            https://docs.python.org/2/reference/lexical_analysis.html#identifiers

        """
        s = string.lower()
        clean_s = s.replace(" ", "_")

        # Make spaces into underscores
        s = re.sub(r"[\s\t\n_]+", "_", s)

        # remove accents, diaeresis and others ó -> o
        s = normalize("NFD", s).encode("ascii", "ignore").decode("utf-8")

        # Remove invalid characters
        s = re.sub(r"[^0-9a-zA-Z_]", "", s)

        # If leading character is not a letter add nvs_.
        # Only letters can be leading characters.
        if prefix is not None:
            s = prefix + "_" + s
        elif re.findall(r"^[0-9]", s) or not s:
            s = "nvs_" + s
        elif re.findall(r"^_", s):
            s = "nvs" + s

        # replace multiple _ after cleaning
        s = re.sub(r"[_]+", "_", s)

        # Check that the string is not a Python identifier
        identifier = s
        i = 1
        while identifier in self._used_words:
            identifier = s + '_' + str(i)
            i += 1

        # include the word in used words to avoid using it againg
        self._used_words.add(identifier)

        if add_to_namespace:
            # include word to the namespace
            self.namespace[string] = identifier
            self.cleanspace[clean_s] = identifier

        return identifier

    def get_original_name(self, identifier):
        """
        Search for the original name of a variable's Python identifier.

        Parameters
        ----------
        identifier: str
            It should be a value in the namespace.

        Rerturns
        --------
        original_name: str
            The original name of the variable.

        """
        for key, value in self.namespace.items():
            if value == identifier:
                return key

        raise ValueError(f"'{identifier}' not found in the namespace.")



from setuptools import setup, find_packages

exec(open('pysd/_version.py').read())
print(__version__)

setup(
    name='pysd',
    version=__version__,
    python_requires='>=3.9',
    author='PySD contributors',
    packages=find_packages(exclude=['docs', 'tests', 'dist', 'build']),
    url='https://github.com/SDXorg/pysd',
    license='LICENSE',
    description='System Dynamics Modeling in Python',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    keywords=['System Dynamics', 'Vensim', 'XMILE'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Intended Audience :: Science/Research',

        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    install_requires=open('requirements.txt').read().strip().split('\n'),
    package_data={
        'translators': [
            '*/parsing_grammars/*.peg'
        ]
    },
    include_package_data=True
)



"""
These are the decorators used by the functions in the model file.
functions.py
"""
from functools import wraps
import inspect


class Cache(object):
    """
    This is the class for the chache. Several cache types can be saved
    in dictionaries and acces using cache.data[cache_type].
    """
    def __init__(self):
        self.cached_funcs = set()
        self.data = {}

    def __call__(self, func, *args):
        """ Decorator for caching """

        @wraps(func)
        def cached_func(*args):
            """ Cache function """
            try:
                return self.data[func.__name__]
            except KeyError:
                value = func(*args)
                self.data[func.__name__] = value
                return value
        return cached_func

    def clean(self):
        """ Cleans the cache """
        self.data = {}


def constant_cache(function, *args):
    """
    Constant cache decorator for all the run
    The original function is saved in 'function' attribuite so we can
    recover it later.
    """
    function.function = function
    function.value = function(*args)

    @wraps(function)
    def wrapper(*args):
        return function.value

    return wrapper



"""
Tools for importing and converting netCDF files generated from simulations run
using PySD.
"""
import itertools
import warnings

from pathlib import Path
from csv import QUOTE_NONE
from typing import Union, Optional

import xarray as xr
import pandas as pd


class NCFile():
    """
    Helper class to extract data from netCDF files.

    Parameters
    ----------
    ncfile: str or pathlib.Path
        Path to the netCDF file to process.
    parallel: bool (optional)
        When True, the Dataset is opened using chunks=-1 (see xarray
        documentation for details) and DataArrays are processed in parallel
        using dask delayed. Dask is not included as a requirement for pysd,
        hence it must be installed separately. Setting parallel=True is
        highly recommended when the Dataset contains large multidimensional
        DataArrays.

    """

    valid_export_file_types = [".csv", ".tab"]

    def __init__(self,
                 filename: Union[str, Path],
                 parallel: Optional[bool] = False) -> None:

        self.ncfile = NCFile._validate_nc_path(filename)
        self.parallel = parallel
        self.ds = self.open_nc()

    def to_text_file(self,
                     outfile: Optional[Union[str, Path]] = "result.tab",
                     subset: Optional[list] = None,
                     time_in_row: Optional[bool] = False,
                     ) -> pd.DataFrame:
        """
        Convert netCDF file contents into comma separated or tab delimited
        file.

        Parameters
        ----------
        outfile: str or pathlib.Path (optional)
            Path to the output file.
        subset: list (optional)
            List of variables to export from the netCDF.
        time_in_row: bool (optional)
            Whether time increases along row.
            Default is False.

        Returns
        -------
        df: pandas.DataFrame
            Dataframe with all colums specified in subset.

        """
        df = self.to_df(subset=subset)

        NCFile.df_to_text_file(df, outfile, time_in_row)

        return df

    def to_df(self,
              subset: Optional[list] = None,
              ) -> pd.DataFrame:
        """
        Wrapper to ds_to_df static method. Convert xarray.Dataset into a
        pandas DataFrame.

        Parameters
        ----------
        subset: list (optional)
            List of variables to export from the Dataset.

        Returns
        -------
        df: pandas.DataFrame
            Dataframe with all colums specified in subset.

        """
        return NCFile.ds_to_df(self.ds, subset, self.parallel)

    def open_nc(self) -> xr.Dataset:
        """
        Loads netCDF file into xarray Dataset. It's basically a wrapper to
        xr.open_dataset to simplify the interface for pysd use case (loading
        simulation results).

        Returns
        -------
        xarray.Dataset

        """
        if self.parallel:
            return xr.open_dataset(self.ncfile, engine="netcdf4", chunks=-1)

        return xr.open_dataset(self.ncfile, engine="netcdf4")

    @staticmethod
    def ds_to_df(ds: xr.Dataset,
                 subset: Optional[list] = None,
                 parallel: Optional[bool] = False,
                 index_dim: Optional[str] = "time"
                 ) -> pd.DataFrame:
        """
        Convert xarray.Dataset into a pandas DataFrame.

        Parameters
        ----------
        ds: xarray.Dataset
            Dataset object.
        subset: list (optional)
            List of variables to export from the Dataset.
        parallel: bool (optional)
            When True, DataArrays are processed in parallel using dask delayed.
            Setting parallel=True is highly recommended when DataArrays are
            large and multidimensional.
        index_dim: str (optional)
            Name of dimensions to use as index of the resulting DataFrame
            (usually "time").

        Returns
        -------
        df: pandas.DataFrame
            Dataframe with all colums specified in subset.

        """
        subset = NCFile._validate_ds_subset(ds, subset)

        if parallel:
            processing_func = NCFile.da_to_dict_delayed
        else:
            processing_func = NCFile.da_to_dict

        savedict = {}

        for name in subset:
            print(f"\nProcessing variable {name}.")
            da = ds[name]
            dims = da.dims

            if not dims or dims == (index_dim,):
                savedict.update({name: da.values.tolist()})
            else:
                savedict.update(processing_func(da, index_dim))

        return NCFile.dict_to_df(savedict)

    @staticmethod
    def df_to_text_file(df: pd.DataFrame, outfile: Path,
                        time_in_row: Optional[bool] = False
                        ) -> None:
        """
        Store pandas DataFrame into csv or tab file.

        Parameters
        ----------
        df: pandas.DataFrame
            DataFrame to save as csv or tab file.
        outfile: str or pathlib.Path
            Path of the output file.
        time_in_row: bool (optional)
            Whether time increases along a column or a row.

        Returns
        -------
        None
        """
        outfile = Path(outfile)

        out_fmt = outfile.suffix

        if out_fmt not in NCFile.valid_export_file_types:
            raise TypeError("Invalid output file format {out_fmt}\n"
                            "Supported formats are csv and tab.")

        outfile.parent. mkdir(parents=True, exist_ok=True)

        if not isinstance(time_in_row, bool):
            raise ValueError("time_in_row argument takes boolen values.")

        # process output file path
        if outfile.suffix == ".csv":
            sep = ","
            df.columns = [col.replace(",", ";") for col in df.columns]
        else:
            sep = "\t"

        if time_in_row:
            df = df.transpose()

        # QUOTE_NONE used to print the csv/tab files as vensim does with
        # special characterse, e.g.: "my-var"[Dimension]
        df.to_csv(outfile, sep=sep, index_label="Time", quoting=QUOTE_NONE)

        print(f"Data saved in '{outfile}'")

    @staticmethod
    def da_to_dict(da: xr.DataArray, index_dim: str) -> dict:
        """
        Splits a DataArray into a dictionary, with keys equal to the name
        of the variable plus all combinations of the cartesian product of
        coordinates within brackets, and values equal to the data
        corresponding to those coordinates along the index_dim dimension.

        Parameters
        ----------
        index_dim: str
            The coordinates of this dimension will not be fixed during
            indexing of the DataArray (i.e. the indexed data will be a
            scalar or an array along this dimension).

        """
        dims, coords = NCFile._get_da_dims_coords(da, index_dim)

        indexes = []
        # TODO: try to achieve the same as itertools.product with
        # xr.DataArray.stack
        for coords_prod in itertools.product(*coords):
            indexes.append(
                NCFile._index_da_by_coord_labels(da, dims, coords_prod)
            )

        return dict(indexes)

    @staticmethod
    def da_to_dict_delayed(da: xr.DataArray, index_dim: str) -> dict:
        """
        Same as da_to_dict, but using dask delayed and compute.
        This function runs much faster when da is a dask array (chunked).

        To use it on its own, you must first make the following imports:

        from dask import delayed, compute
        from dask.diagnostics import ProgressBar

        Parameters
        ----------
        index_dim: str
            The coordinates of this dimension will not be fixed during
            indexing (the indexed data will be an array along this dimension).

        """
        namespace = dir()
        if not all(
             map(lambda x: x in namespace, [
                "delayed", "compute", "ProgressBar"]
                )):
            from dask import delayed, compute
            from dask.diagnostics import ProgressBar

        dims, coords = NCFile._get_da_dims_coords(da, index_dim)

        # loading data into memory for faster indexing
        da.load()

        indexes = []
        # TODO: try to achieve the same as itertools.product with
        # xr.DataArray.stack
        for coords_prod in itertools.product(*coords):
            x = delayed(
                NCFile._index_da_by_coord_labels
            )(da, dims, coords_prod)
            indexes.append(x)

        with ProgressBar():
            res = compute(*indexes)

        return dict(res)

    @staticmethod
    def dict_to_df(d: dict) -> pd.DataFrame:
        """
        Convert a dict to a pandas Dataframe.

        Parameters
        ----------
        d: dict
            Dictionary to convert to pandas DataFrame.

        """
        if "time" not in d:
            raise KeyError("Missing time key.")

        return pd.DataFrame(d).set_index('time')

    @staticmethod
    def _validate_nc_path(nc_path: Union[str, Path]) -> Path:
        """
        Checks validity of the nc_path passed by the user. We run these
        checks because xarray Exceptions are not very explicit.
        """

        if not isinstance(nc_path, (str, Path)):
            raise TypeError(f"Invalid file path type: {type(nc_path)}.\n"
                            "Please provide string or pathlib Path")

        nc_path = Path(nc_path)

        if not nc_path.is_file():
            raise FileNotFoundError(f"{nc_path} could not be found.")

        if not nc_path.suffix == ".nc":
            raise ValueError("Input file must have nc extension.")

        return nc_path

    @staticmethod
    def _validate_ds_subset(ds: xr.Dataset, subset: list) -> list:
        """
        If subset=None, it returns a list with all variable names in the ds.
        If var names in subset are present in ds, it returns them, else it
        warns the user.

        Parameters
        ----------
        subset: list
            Subset of variable names in the xarray Dataset.

        """
        # use all variable names
        if not subset:
            new_subset = [name for name in ds.data_vars.keys()]
        else:
            if not isinstance(subset, list) or \
                 not all(map(lambda x: isinstance(x, str), subset)):
                raise TypeError("Subset argument must be a list of strings.")

            new_subset = []
            for name in subset:
                if name in ds.data_vars.keys():
                    new_subset.append(name)
                else:
                    warnings.warn(f"{name} not in Dataset.")

            if not new_subset:
                raise ValueError("None of the elements of the subset are "
                                 "present in the Dataset.")

        # adding time in the final subset
        if "time" not in new_subset:
            new_subset.append("time")

        return new_subset

    @staticmethod
    def _index_da_by_coord_labels(da: xr.DataArray, dims: list,
                                  coords: tuple) -> tuple:
        """
        Generates variable names, combining the actual name of the variable
        with the coordinate names between brackets and separated by commas,
        and indexes the DataArray by the coordinate names specified in
        the coords argument.

        Parameters
        ----------
        da: xr.Dataset
            Dataset to be indexed.
        dims: list
            Dimensions along which the DataArray will be indexed.
        coords: tuple
            Coordinate names for each of the dimensons in the dims list.

        Returns
        -------
        A tuple consisting of the string
        var_name[dim_1_coord_j, ..., dim_n_coord_k] in the first index, and the
        indexed data as the second index.

        """
        name = da.name
        idx = dict(zip(dims, coords))
        subs = "[" + ",".join(map(lambda x: str(x), coords)) + "]"

        return name + subs, da.loc[idx].values

    @staticmethod
    def _get_da_dims_coords(da: xr.DataArray, exclude_dim: str) -> tuple:
        """
        Returns the dimension names and coordinate labels in two
        separate lists. If a dimension name is in the exclude_dims
        list, the returned dims and coords will not include it.

        Parameters
        ----------
        exclude_dim: str
            Names of dimension to exclude.

        Returns
        -------
        dims: list
            List containing the names of the DataArray dimensions.
        coords: list
            List of lists of coordinates for each dimension.

        """
        dims, coords = [], []

        for dim in da.dims:
            if dim != exclude_dim:
                dims.append(dim)
                coords.append(da.coords[dim].values)

        return dims, coords



"""
The main Abstract dataclasses provide the structure for the information
from the Component level to the Model level. This classes are hierarchical
An AbstractComponent will be inside an AbstractElement, which is inside an
AbstractSection, which is a part of an AbstractModel.

"""
from dataclasses import dataclass
from typing import Tuple, List, Union
from pathlib import Path


@dataclass
class AbstractComponent:
    """
    Dataclass for a regular component.

    Parameters
    ----------
    subscripts: tuple
        Tuple of length two with first argument the list of subscripts
        in the variable definition and the second argument the list of
        subscripts list that must be ignored (EXCEPT).
    ast: object
        The AbstractSyntaxTree of the component expression
    type: str (optional)
        The type of component. 'Auxiliary' by default.
    subtype: str (optional)
        The subtype of component. 'Normal' by default.

    """
    subscripts: Tuple[List[str], List[List[str]]]
    ast: object
    type: str = "Auxiliary"
    subtype: str = "Normal"

    def __str__(self) -> str:  # pragma: no cover
        return "AbstractComponent %s\n" % (
            "%s" % repr(list(self.subscripts)) if self.subscripts else "")

    def dump(self, depth=None, indent="") -> str:  # pragma: no cover
        """
        Dump the component to a printable version.

        Parameters
        ----------
        depth: int (optional)
            The number of depht levels to show in the dumped output.
            Default is None which will dump everything.

        indent: str (optional)
            The indent to use for a lower level object. Default is ''.

        """
        if depth == 0:
            return self.__str__()

        return self.__str__() + "\n" + self._str_child(depth, indent)

    def _str_child(self, depth, indent) -> str:  # pragma: no cover
        return str(self.ast).replace("\t", indent).replace("\n", "\n" + indent)


@dataclass
class AbstractUnchangeableConstant(AbstractComponent):
    """
    Dataclass for an unchangeable constant component. This class is a child
    of AbstractComponent.

    Parameters
    ----------
    subscripts: tuple
        Tuple of length two with first argument the list of subscripts
        in the variable definition and the second argument the list of
        subscripts list that must be ignored (EXCEPT).
    ast: object
        The AbstractSyntaxTree of the component expression
    type: str (optional)
        The type of component. 'Constant' by default.
    subtype: str (optional)
        The subtype of component. 'Unchangeable' by default.

    """
    subscripts: Tuple[List[str], List[List[str]]]
    ast: object
    type: str = "Constant"
    subtype: str = "Unchangeable"

    def __str__(self) -> str:  # pragma: no cover
        return "AbstractUnchangeableConstant %s\n" % (
            "%s" % repr(list(self.subscripts)) if self.subscripts else "")


@dataclass
class AbstractLookup(AbstractComponent):
    """
    Dataclass for a lookup component. This class is a child of
    AbstractComponent.

    Parameters
    ----------
    subscripts: tuple
        Tuple of length two with first argument the list of subscripts
        in the variable definition and the second argument the list of
        subscripts list that must be ignored (EXCEPT).
    ast: object
        The AbstractSyntaxTree of the component expression
    arguments: str (optional)
        The name of the argument to use. 'x' by default.
    type: str (optional)
        The type of component. 'Lookup' by default.
    subtype: str (optional)
        The subtype of component. 'Hardcoded' by default.

    """
    subscripts: Tuple[List[str], List[List[str]]]
    ast: object
    arguments: str = "x"
    type: str = "Lookup"
    subtype: str = "Hardcoded"

    def __str__(self) -> str:  # pragma: no cover
        return "AbstractLookup %s\n" % (
            "%s" % repr(list(self.subscripts)) if self.subscripts else "")


@dataclass
class AbstractData(AbstractComponent):
    """
    Dataclass for a data component. This class is a child
    of AbstractComponent.

    Parameters
    ----------
    subscripts: tuple
        Tuple of length two with first argument the list of subscripts
        in the variable definition and the second argument the list of
        subscripts list that must be ignored (EXCEPT).
    ast: object
        The AbstractSyntaxTree of the component expression
    keyword: str or None (optional)
        The data object keyword ('interpolate', 'hold_backward',
        'look_forward', 'raw'). Default is None.
    type: str (optional)
        The type of component. 'Data' by default.
    subtype: str (optional)
        The subtype of component. 'Normal' by default.

    """
    subscripts: Tuple[List[str], List[List[str]]]
    ast: object
    keyword: Union[str, None] = None
    type: str = "Data"
    subtype: str = "Normal"

    def __str__(self) -> str:  # pragma: no cover
        return "AbstractData (%s)  %s\n" % (
            self.keyword,
            "%s" % repr(list(self.subscripts)) if self.subscripts else "")

    def dump(self, depth=None, indent="") -> str:  # pragma: no cover
        """
        Dump the component to a printable version.

        Parameters
        ----------
        depth: int (optional)
            The number of depht levels to show in the dumped output.
            Default is None which will dump everything.

        indent: str (optional)
            The indent to use for a lower level object. Default is ''.

        """
        if depth == 0:
            return self.__str__()

        return self.__str__() + "\n" + self._str_child(depth, indent)

    def _str_child(self, depth, indent) -> str:  # pragma: no cover
        return str(self.ast).replace("\n", "\n" + indent)


@dataclass
class AbstractElement:
    """
    Dataclass for an element.

    Parameters
    ----------
    name: str
        The name of the element.
    components: list
        The list of AbstractComponents that define this element.
    units: str (optional)
        The units of the element. '' by default.
    limits: tuple (optional)
        The limits of the element. (None, None) by default.
    units: str (optional)
        The documentation of the element. '' by default.

    """
    name: str
    components: List[AbstractComponent]
    units: str = ""
    limits: tuple = (None, None)
    documentation: str = ""

    def __str__(self) -> str:  # pragma: no cover
        return "AbstractElement:\t%s (%s, %s)\n%s\n" % (
            self.name, self.units, self.limits, self.documentation)

    def dump(self, depth=None, indent="") -> str:  # pragma: no cover
        """
        Dump the element to a printable version.

        Parameters
        ----------
        depth: int (optional)
            The number of depht levels to show in the dumped output.
            Default is None which will dump everything.

        indent: str (optional)
            The indent to use for a lower level object. Default is ''.

        """
        if depth == 0:
            return self.__str__()
        elif depth is not None:
            depth -= 1

        return self.__str__() + "\n" + self._str_child(depth, indent)

    def _str_child(self, depth, indent) -> str:  # pragma: no cover
        return "\n".join([
            component.dump(depth, indent) for component in self.components
            ]).replace("\n", "\n" + indent)


@dataclass
class AbstractControlElement(AbstractElement):
    """
    Dataclass for a control element. This class is a child of
    AbstractElement and has the same attributes.

    Parameters
    ----------
    name: str
        The name of the element.
    components: list
        The list of AbstractComponents that define this element.
    units: str (optional)
        The units of the element. '' by default.
    limits: tuple (optional)
        The limits of the element. (None, None) by default.
    units: str (optional)
        The documentation of the element. '' by default.

    """
    name: str
    components: List[AbstractComponent]
    units: str = ""
    limits: tuple = (None, None)
    documentation: str = ""

    def __str__(self) -> str:  # pragma: no cover
        return "AbstractControlElement:\t%s (%s, %s)\n%s\n" % (
            self.name, self.units, self.limits, self.documentation)


@dataclass
class AbstractSubscriptRange:
    """
    Dataclass for a subscript range.

    Parameters
    ----------
    name: str
        The name of the element.
    subscripts: list or str or dict
        The subscripts as a list of strings for a regular definition,
        str for a copy definition and as a dict for a GET XLS/DIRECT
        definition.
    mapping: list
        The list of subscript range that can be mapped to.

    """
    name: str
    subscripts: Union[list, str, dict]
    mapping: list

    def __str__(self) -> str:  # pragma: no cover
        return "AbstractSubscriptRange:\t%s\n\t%s\n" % (
            self.name,
            "%s <- %s" % (self.subscripts, self.mapping)
            if self.mapping else self.subscripts)

    def dump(self, depth=None, indent="") -> str:  # pragma: no cover
        """
        Dump the subscript range to a printable version.

        Parameters
        ----------
        depth: int (optional)
            The number of depht levels to show in the dumped output.
            Default is None which will dump everything.

        indent: str (optional)
            The indent to use for a lower level object. Default is ''.

        """
        return self.__str__()


@dataclass
class AbstractConstraint:
    """
    Dataclass for a constraint

    Parameters
    ----------
    name: str
        Name of the constraint variable.
    subscripts: list or str or dict
        The subscripts as a list of strings for a regular definition,
        str for a copy definition and as a dict for a GET XLS/DIRECT
        definition.
    expression: str
        Unparsed constraint expression.
    """

    name: str
    subscripts: Union[list, str, dict]
    expression: str

    def __str__(self) -> str:  # pragma: no cover
        raise NotImplementedError()


@dataclass
class AbstractTestInput:
    """
    Dataclass for a test inputs

    Parameters
    ----------
    name: str
        Name of the test inputs variable.
    subscripts: list or str or dict
        The subscripts as a list of strings for a regular definition,
        str for a copy definition and as a dict for a GET XLS/DIRECT
        definition.
    expression: str
        Unparsed test inputs expression.
    """
    name: str
    subscripts: Union[list, str, dict]
    expression: str

    def __str__(self) -> str:  # pragma: no cover
        raise NotImplementedError()


@dataclass
class AbstractSection:
    """
    Dataclass for a section.

    Parameters
    ----------
    name: str
        Section name. '__main__' for the main section or the macro name.
    path: pathlib.Path
        Section path. It should be the model name for main  section and
        the clean macro name for a macro.
    section_type: str ('main' or 'macro')
        The section type.
    params: list
        List of params that takes the section. In the case of main
        section it will be an empty list.
    returns: list
        List of variables that returns the section. In the case of main
        section it will be an empty list.
    subscripts: tuple
        Tuple of AbstractSubscriptRanges that are defined in the section.
    elements: tuple
        Tuple of AbstractElements that are defined in the section.
    constraints: tuple
        Tuple of AbstractConstraints that are defined in the section.
    test_inputs: tuple
        Tuple of TestInputs that are defined in the section.
    split: bool
        If split is True the created section will split the variables
        depending on the views_dict.
    views_dict: dict
        The dictionary of the views. Giving the variables classified at
        any level in order to split them by files.

    """
    name: str
    path: Path
    type: str  # main, macro or module
    params: List[str]
    returns: List[str]
    subscripts: Tuple[AbstractSubscriptRange]
    elements: Tuple[AbstractElement]
    constraints: Tuple[AbstractConstraint]
    test_inputs: Tuple[AbstractTestInput]
    split: bool
    views_dict: Union[dict, None]

    def __str__(self) -> str:  # pragma: no cover
        return "AbstractSection (%s):\t%s (%s)\n" % (
            self.type, self.name, self.path)

    def dump(self, depth=None, indent="") -> str:  # pragma: no cover
        """
        Dump the section to a printable version.

        Parameters
        ----------
        depth: int (optional)
            The number of depht levels to show in the dumped output.
            Default is None which will dump everything.

        indent: str (optional)
            The indent to use for a lower level object. Default is ''.

        """
        if depth == 0:
            return self.__str__()
        elif depth is not None:
            depth -= 1

        return self.__str__() + "\n" + self._str_child(depth, indent)

    def _str_child(self, depth, indent) -> str:  # pragma: no cover
        return "\n".join([
            element.dump(depth, indent) for element in self.subscripts
            ] + [
            element.dump(depth, indent) for element in self.elements
            ]).replace("\n", "\n" + indent)


@dataclass
class AbstractModel:
    """
    Dataclass for a model.

    Parameters
    ----------
    original_path: pathlib.Path
        The path to the original file.
    sections: tuple
        Tuple of AbstractSectionss that are defined in the model.

    """
    original_path: Path
    sections: Tuple[AbstractSection]

    def __str__(self) -> str:  # pragma: no cover
        return "AbstractModel:\t%s\n" % self.original_path

    def dump(self, depth=None, indent="") -> str:  # pragma: no cover
        """
        Dump the model to a printable version.

        Parameters
        ----------
        depth: int (optional)
            The number of depht levels to show in the dumped output.
            Default is None which will dump everything.

        indent: str (optional)
            The indent to use for a lower level object. Default is ''.

        """
        if depth == 0:
            return self.__str__()
        elif depth is not None:
            depth -= 1

        return self.__str__() + "\n" + self._str_child(depth, indent)

    def _str_child(self, depth, indent) -> str:  # pragma: no cover
        return "\n".join([
            section.dump(depth, indent) for section in self.sections
            ]).replace("\n", "\n" + indent)



"""
The Section class allows parsing a model section into Elements using the
SectionElementsVisitor. The final result can be exported to an
AbstractSection class in order to build a model in another programming
language.
A section is either the main model (without the macros), or a
macro definition.
"""
from typing import List, Union
from pathlib import Path
import parsimonious

from ..structures.abstract_model import\
    AbstractElement, AbstractControlElement, AbstractSection

from . import vensim_utils as vu
from .vensim_element import Element, SubscriptRange, Component,\
                            Constraint, TestInput


class Section():
    """
    The Section class allows parsing the elements of a model section.

    Parameters
    ----------
    name: str
        Section name. That is, '__main__' for the main section, and the
        macro name for macros.

    path: pathlib.Path
        Section path. The model name for the main section and the clean
        macro name for a macro.

    section_type: str ('main' or 'macro')
        The section type.

    params: list
        List of parameters that the section takes. If it is the main
        section, it will be an empty list.

    returns: list
        List of variables that returns the section. If it is the main
        section, it will be an empty list.

    content: str
        Section content as string.

    split: bool
        If True, the created section will split the variables
        depending on the views_dict.

    views_dict: dict
        The dictionary of the views. Giving the variables classified at
        any level in order to split them by files.

    """
    def __init__(self, name: str, path: Path, section_type: str,
                 params: List[str], returns: List[str],
                 content: str, split: bool, views_dict: Union[dict, None]):
        self.name = name
        self.path = path
        self.type = section_type
        self.params = params
        self.returns = returns
        self.content = content
        self.split = split
        self.views_dict = views_dict
        self.elements = None
        self.components = None
        self.subscripts = None

    def __str__(self):  # pragma: no cover
        return "\nSection: %s\n" % self.name

    @property
    def _verbose(self) -> str:  # pragma: no cover
        """Get section information."""
        text = self.__str__()
        if self.elements:
            for element in self.elements:
                text += element._verbose
        else:
            text += self.content

        return text

    @property
    def verbose(self):  # pragma: no cover
        """Print section information to standard output."""
        print(self._verbose)

    def parse(self, parse_all: bool = True) -> None:
        """
        Parse section object with parsimonious using the grammar given in
        'parsing_grammars/section_elements.peg' and the class
        SectionElementsVisitor to visit the parsed expressions.

        This will break the section (__main__ or macro) in VensimElements,
        which are each model expression LHS and RHS with already parsed
        units and description.

        Parameters
        ----------
        parse_all: bool (optional)
            If True, the created VensimElement objects will be
            automatically parsed. Otherwise, these objects will only be
            added to self.elements but not parsed. Default is True.

        """
        # parse the section to get the elements
        tree = vu.Grammar.get("section_elements").parse(self.content)
        self.elements = SectionElementsVisitor(tree).entries

        if parse_all:
            # parse all elements
            self.elements = [element.parse() for element in self.elements]

            # split subscripts and reality checks from other components
            self.subscripts = [
                element for element in self.elements
                if isinstance(element, SubscriptRange)
            ]
            self.components = [
                element for element in self.elements
                if isinstance(element, Component)
            ]
            self.constraints = [
                element for element in self.elements
                if isinstance(element, Constraint)
            ]
            self.test_inputs = [
                element for element in self.elements
                if isinstance(element, TestInput)
            ]

            # reorder element list for better printing
            self.elements = self.subscripts + self.components\
                + self.constraints + self.test_inputs

            [component.parse() for component in self.components]
            [component.parse() for component in self.constraints]
            [component.parse() for component in self.test_inputs]

    def get_abstract_section(self) -> AbstractSection:
        """
        Instantiate an object of the AbstractSection class that will be used
        during the building process. This method should be called after
        parsing the section (self.parse). This method is automatically called
        by Model's get_abstract_model method, and automatically generates the
        AbstractSubscript ranges and merges the components in elements. It also
        calls the get_abstract_components method from each model component.

        Returns
        -------
        AbstractSection: AbstractSection
          AbstractSection object that can be used for building the model
          in another programming language.

        """
        return AbstractSection(
            name=self.name,
            path=self.path,
            type=self.type,
            params=self.params,
            returns=self.returns,
            subscripts=[
                subs_range.get_abstract_subscript_range()
                for subs_range in self.subscripts
            ],
            elements=self._merge_components(),
            constraints=[
                constraint.get_abstract_component()
                for constraint in self.constraints],
            test_inputs=[
                test_input.get_abstract_component()
                for test_input in self.test_inputs],
            split=self.split,
            views_dict=self.views_dict
        )

    def _merge_components(self) -> List[AbstractElement]:
        """Merge model components by their name."""
        control_vars = ["initial_time", "final_time", "time_step", "saveper"]
        merged = {}
        for component in self.components:
            # get a safe name to merge (case and white/underscore sensitivity)
            name = component.name.lower().replace(" ", "_")
            if name not in merged:
                # create new element if it is the first component
                if name in control_vars:
                    merged[name] = AbstractControlElement(
                        name=component.name,
                        components=[])
                else:
                    merged[name] = AbstractElement(
                        name=component.name,
                        components=[])

            if component.units:
                # add units to element data
                merged[name].units = component.units
            if component.limits:
                # add limits to element data
                merged[name].limits = component.limits
            if component.documentation:
                # add documentation to element data
                merged[name].documentation = component.documentation

            # add AbstractComponent to the list of components
            merged[name].components.append(component.get_abstract_component())

        return list(merged.values())


class SectionElementsVisitor(parsimonious.NodeVisitor):
    """
    Visit section elements to get their equation, units and documentation.
    """
    # TODO include units parsing
    def __init__(self, ast):
        self.entries = []
        self.visit(ast)

    def visit_entry(self, n, vc):
        self.entries.append(
            Element(
                equation=vc[0].strip(),
                units=vc[2].strip(),
                documentation=vc[4].strip(),
            )
        )

    def generic_visit(self, n, vc):
        return "".join(filter(None, vc)) or n.text or ""
